@@ -55,6 +55,10 @@ else:
 
 import numpy as np
 
+# Known ENVI data file extensions. Upper and lower case versions will be
+# recognized, as well as interleaves ('bil', 'bip', 'bsq'), and no extension.
+KNOWN_EXTS = ['img', 'dat', 'sli', 'hyspex', 'raw']
+
 dtype_map = [('1', np.uint8),                   # unsigned byte
              ('2', np.int16),                   # 16-bit int
              ('3', np.int32),                   # 32-bit int
@@ -69,12 +73,15 @@ dtype_map = [('1', np.uint8),                   # unsigned byte
 envi_to_dtype = dict((k, np.dtype(v).char) for (k, v) in dtype_map)
 dtype_to_envi = dict(tuple(reversed(item)) for item in list(envi_to_dtype.items()))
 
-class EnviException(Exception):
+from spectral import SpyException
+from .spyfile import FileNotFoundError, InvalidFileError
+
+class EnviException(SpyException):
     '''Base class for ENVI file-related exceptions.'''
     pass
 
-class EnviDataTypeError(TypeError):
-    '''Exception raised when saving invalid image data type to ENVI format.
+class EnviDataTypeError(EnviException, TypeError):
+    '''Raised when saving invalid image data type to ENVI format.
     '''
     def __init__(self, dtype):
         msg = 'Image data type "{0}" can not be saved to ENVI data file. ' \
@@ -82,11 +89,11 @@ class EnviDataTypeError(TypeError):
           'data type names.'.format(np.dtype(dtype).name)
         super(EnviDataTypeError, self).__init__(msg)
 
-class EnviFeatureNotSupported(NotImplementedError):
+class EnviFeatureNotSupported(EnviException, NotImplementedError):
     '''A specified ENVI capability is not supported by the spectral module.'''
     pass
 
-class FileNotAnEnviHeader(EnviException):
+class FileNotAnEnviHeader(EnviException, InvalidFileError):
     '''Raised when "ENVI" does not appear on the first line of the file.'''
     def __init__(self, msg):
         super(FileNotAnEnviHeader, self).__init__(msg)
@@ -97,12 +104,16 @@ class MissingEnviHeaderParameter(EnviException):
         msg = 'Mandatory parameter "%s" missing from header file.' % param
         super(MissingEnviHeaderParameter, self).__init__(msg)
 
-class EnviHeaderParsingError(EnviException):
+class EnviHeaderParsingError(EnviException, InvalidFileError):
     '''Raised upon failure to parse parameter/value pairs from a file.'''
     def __init__(self):
         msg = 'Failed to parse ENVI header file.'
         super(EnviHeaderParsingError, self).__init__(msg)
 
+class EnviDataFileNotFoundError(EnviException, FileNotFoundError):
+    '''Raised when data file associated with a header is not found.'''
+    pass
+
 def _validate_dtype(dtype):
     '''Raises EnviDataTypeError if dtype can not be written to ENVI file.'''
     typename = np.dtype(dtype).name
@@ -121,6 +132,8 @@ def read_envi_header(file):
     dictionary as strings.  Header field names are treated as case
     insensitive and all keys in the dictionary are lowercase.
     '''
+    import warnings
+    from spectral import settings
     f = builtins.open(file, 'r')
 
     try:
@@ -141,6 +154,8 @@ def read_envi_header(file):
     f.close()
 
     dict = {}
+    have_nonlowercase_param = False
+    support_nonlowercase_params = settings.envi_support_nonlowercase_params
     try:
         while lines:
             line = lines.pop(0)
@@ -148,7 +163,11 @@ def read_envi_header(file):
             if line[0] == ';': continue
 
             (key, sep, val) = line.partition('=')
-            key = key.strip().lower()
+            key = key.strip()
+            if not key.islower():
+                have_nonlowercase_param = True
+                if not support_nonlowercase_params:
+                    key = key.lower()
             val = val.strip()
             if val and val[0] == '{':
                 str = val.strip()
@@ -167,6 +186,14 @@ def read_envi_header(file):
             else:
                 dict[key] = val
 
+        if have_nonlowercase_param and not support_nonlowercase_params:
+            msg = 'Parameters with non-lowercase names encountered ' \
+                  'and converted to lowercase. To retain source file ' \
+                  'parameter name capitalization, set ' \
+                  'spectral.setttings.envi_support_nonlowercase_params to ' \
+                  'True.'
+            warnings.warn(msg)
+            print('Header parameter names converted to lower case.')
         return dict
     except:
         raise EnviHeaderParsingError()
@@ -273,7 +300,7 @@ def open(file, image=None):
 
     Raises:
 
-        TypeError, IOError.
+        TypeError, EnviDataFileNotFoundError
 
     If the specified file is not found in the current directory, all
     directories listed in the SPECTRAL_DATA environment variable will be
@@ -289,8 +316,8 @@ def open(file, image=None):
     import numpy
     import spectral
 
-    headerPath = find_file_path(file)
-    h = read_envi_header(headerPath)
+    header_path = find_file_path(file)
+    h = read_envi_header(header_path)
     check_compatibility(h)
     p = gen_params(h)
 
@@ -299,21 +326,23 @@ def open(file, image=None):
     #  Validate image file name
     if not image:
         #  Try to determine the name of the image file
-        headerDir = os.path.split(headerPath)
-        if headerPath[-4:].lower() == '.hdr':
-            headerPathTitle = headerPath[:-4]
-            exts = ['', 'img', 'IMG', 'dat', 'DAT', 'sli', 'SLI', 'hyspex'] +\
-                   [inter.lower(), inter.upper()]
+        (header_path_title, header_ext) = os.path.splitext(header_path)
+        if header_ext.lower() == '.hdr':
+            exts = [ext.lower() for ext in KNOWN_EXTS] + [inter.lower()]
+            exts = [''] + exts + [ext.upper() for ext in exts]
             for ext in exts:
                 if len(ext) == 0:
-                    testname = headerPathTitle
+                    testname = header_path_title
                 else:
-                    testname = headerPathTitle + '.' + ext
+                    testname = header_path_title + '.' + ext
                 if os.path.isfile(testname):
                     image = testname
                     break
         if not image:
-            raise IOError('Unable to determine image file name.')
+            msg = 'Unable to determine the ENVI data file name for the ' \
+              'given header file. You can specify the data file by passing ' \
+              'its name as the optional `image` argument to envi.open.'
+            raise EnviDataFileNotFoundError(msg)
     else:
         image = find_file_path(image)
 
@@ -353,6 +382,11 @@ def open(file, image=None):
             pass
     img.bands.band_unit = h.get('wavelength units', None)
 
+    if 'bbl' in h:
+        try:
+            h['bbl'] = [int(b) for b in h['bbl']]
+        except:
+            print('Unable to parse bad band list (bbl) in header as integers.')
     return img
 
 
@@ -367,15 +401,15 @@ def check_new_filename(hdr_file, img_ext, force):
     hdr_file = os.path.realpath(hdr_file)
     (base, ext) = os.path.splitext(hdr_file)
     if ext.lower() != '.hdr':
-        raise ValueError('Header file name must end in ".hdr" or ".HDR".')
+        raise EnviException('Header file name must end in ".hdr" or ".HDR".')
     image_file = base + img_ext
     if not force:
         if os.path.isfile(hdr_file):
-            raise Exception('Header file %s already exists. Use `force` '
-                            'keyword to force overwrite.' % hdr_file)
+            raise EnviException('Header file %s already exists. Use `force` '
+                                'keyword to force overwrite.' % hdr_file)
         if os.path.isfile(image_file):
-            raise Exception('Image file %s already exists. Use `force` '
-                            'keyword to force overwrite.' % image_file)
+            raise EnviException('Image file %s already exists. Use `force` '
+                                'keyword to force overwrite.' % image_file)
     return (hdr_file, image_file)
 
 
@@ -586,6 +620,7 @@ def _prepared_data_and_metadata(hdr_file, image, **kwargs):
         if len(data.shape) == 2:
             data = data[:, :, np.newaxis]
         swap = False
+        metadata = {}    
     elif isinstance(image, SpyFile):
         if image.using_memmap is True:
             data = image._memmap
@@ -596,12 +631,17 @@ def _prepared_data_and_metadata(hdr_file, image, **kwargs):
             data = image.load(dtype=image.dtype, scale=False)
             src_interleave = 'bip'
             swap = False
+        metadata = image.metadata.copy()
     else:
-        data = image.load()
+        data = image.load(scale=False)
         src_interleave = 'bip'
         swap = False
+        if hasattr(image, 'metadata'):
+            metadata = image.metadata.copy()
+        else:
+            metadata = {}
 
-    metadata = kwargs.get('metadata', {}).copy()
+    metadata.update(kwargs.get('metadata', {}))
     add_image_info_to_metadata(image, metadata)
     if hasattr(image, 'bands'):
         add_band_info_to_metadata(image.bands, metadata)
@@ -818,15 +858,15 @@ def create_image(hdr_file, metadata=None, **kwargs):
 
     # Verify minimal set of parameters have been provided
     if 'lines' not in metadata:
-        raise Exception('Number of image rows is not defined.')
+        raise EnviException('Number of image rows is not defined.')
     elif 'samples' not in metadata:
-        raise Exception('Number of image columns is not defined.')
+        raise EnviException('Number of image columns is not defined.')
     elif 'bands' not in metadata:
-        raise Exception('Number of image bands is not defined.')
+        raise EnviException('Number of image bands is not defined.')
     elif 'samples' not in metadata:
-        raise Exception('Number of image columns is not defined.')
+        raise EnviException('Number of image columns is not defined.')
     elif 'data type' not in metadata:
-        raise Exception('Image data type is not defined.')
+        raise EnviException('Image data type is not defined.')
 
     params = gen_params(metadata)
     dt = np.dtype(params.dtype).char
